# FILE: commandline.tcl
#

Oc_Class Oc_CommandLine {

    # Might want to make this a public common when Oc_Classes have a
    # boilerplate Configure proc.  For now, use the Switch proc.
    common switch -

    array common body
    array common valid
    array common arg
    array common doc

    ClassConstructor {
	$class Option help {} {
	    Oc_Log Log [Oc_CommandLine Usage] info
	    exit
	} "Display this help message and exit"
	$class Option version {} {
	    Oc_Log Log "[Oc_Main GetAppName] [Oc_Main GetVersion]" info
	    exit
	} "Display version information and exit"
	$class Option $switch {} {#do nothing} "End of options"
    }

    proc Panic {msg} {
	Oc_Log Log $msg\n[$class Usage] panic $class
	exit 1
    }

    proc Option {name argList script {docString ""}} {
	if {[string length $script] == 0} {
	    catch {unset arg($name) body($name) doc($name) valid($name)}
	    return
	}
	if {[catch {set argList [lrange $argList 0 end]}]} {
	    return -code error "Argument list for $name is not a valid list"
	}
	set docList [list $docString]
	set validList [list]
	set argSpec [list]
	foreach triple $argList {
	    if {[catch {set triple [lrange $triple 0 end]}]} {
	        return -code error \
			"Argument '$triple' for $name is not a valid list"
	    }
	    lappend argSpec [lindex $triple 0]
	    set validBody [string trim [lindex $triple 1]]
	    if {[string length $validBody]} {
		lappend validList $validBody
	    } else {
		lappend validList {return 1}
	    }
	    set docString [string trim [lindex $triple 2]]
	    if {[string compare $switch $name]} {
		if {[string length $docString]} {
		    lappend docList "where [lindex $triple 0] $docString"
		}
	    } else {
		lappend docList $docString
	    }
	}
	set arg($name) $argSpec
	set body($name) $script
	set doc($name) $docList
	set valid($name) $validList
    }

    proc Parse {argv} {
	set rest [list]
	set order [list]
	# In case option handling tries to redefine $switch
	set sswitch $switch
	while {[llength $argv]} {
	    set option [lindex $argv 0]
	    set argv [lrange $argv 1 end]
	    if {[string first $sswitch $option]} {
		# Not an option switch
		lappend rest $option
	    } else {
		# An option switch.  Strip leading switch characters
		set option [string range $option [string length $sswitch] end]
		if {[string compare $sswitch $option] == 0} {
		    # Double switch ==> End of option processing
		    break
		}
		set matches [array names body $option*]
		if {[llength $matches] > 1} {
		    # Exact match overrides ambiguity
		    if {[info exists body($option)]} {
			set matches [list $option]
		    } else {
			$class Panic "Ambiguous option $sswitch$option matches\
				$sswitch[join $matches ", $sswitch"]"
		    }
		}
		if {[llength $matches] == 0} {
		    $class Panic "Unknown option: $sswitch$option"
		}
		set match [lindex $matches 0]
		set numargs [llength $arg($match)]
		set optargs [lrange $argv 0 [expr {$numargs - 1}]]
		if {[llength $argv] < $numargs} {
		    $class Panic "Argument shortage for $sswitch$match"
		}
		set argv [lrange $argv $numargs end]
		# Call validation scripts
		foreach a $arg($match) v $valid($match) o $optargs {
		    proc _${class}_Validate $arg($match) $v
		    set code [catch {eval _${class}_Validate $optargs} result]
		    rename _${class}_Validate {}
		    if {$code} {
			global errorInfo
			error "Programming error\nError in validation script\
				for '$sswitch$match $a'\n\t$result" $errorInfo
		    }
		    if {[catch {expr {$result && $result}}]} {
			set msg "Programming error\nValidation script return\
				for '$sswitch$match $a' is not 0 or 1."
			error $msg $msg
		    }
		    if {!$result} {
			$class Panic "Invalid argument $a=$o\
				 in\n\t$sswitch$match $arg($match)"
		    }
		}
		# Arguments are valid.  Save for later evaluation
		if {[info exists save($match)]} {
		    set idx [lsearch -exact $order $match]
		    set order [lreplace $order $idx $idx]
		}
		set save($match) $optargs
		lappend order $match
	    }
	}
	# Step through options in order
	foreach opt $order {
	    proc _${class}_HandleOption $arg($opt) $body($opt)
	    set code [catch {eval _${class}_HandleOption $save($opt)} result]
	    rename _${class}_HandleOption {}
	    if {$code} {
		error $result "\n$result\n   \
			(command line option '$sswitch$opt')"
	    }
	}
	# Collect all remaining arguments into $rest
	eval lappend rest $argv
	# Create argument specification for procs
	set argSpec [list]
	set bodyPrefix ""
	set optcount 0
	set min 0
	foreach namePair $arg($sswitch) {
	    set name [lindex $namePair 0]
	    set type [lindex $namePair 1]
	    switch -- $type {
		optional {
		    lappend argSpec [list $name {}]
		    incr optcount
		}
		list {
		    lappend argSpec args
		    set bodyPrefix "set $name \$args;"
		    incr optcount
		    break
		}
		default {
		    if {$optcount} {
			set msg "Programming error\nRequired argument $name\
				follows optional argument(s)."
			error $msg $msg
		    }
		    lappend argSpec $name
		    incr min
		}
	    }
	}
	if {$optcount + $min < [llength $arg($sswitch)]} {
	    set msg "Programming error\nList argument $name must be last"
	    error $msg $msg
	}
	if {[llength $rest] < $min} {
	    $class Panic "Missing required argument:\
		    [lindex [lindex $arg($sswitch) [llength $rest]] 0]"
	}
	if {[string compare args [lindex $argSpec end]] 
		&& [llength $rest] > $min + $optcount} {
	    $class Panic "Too many arguments"
	}
	# Run the validation scripts
	foreach namePair $arg($sswitch) v $valid($sswitch) o $rest {
	    set name [lindex $namePair 0]
	    proc _${class}_Validate $argSpec $bodyPrefix$v
	    set code [catch {eval _${class}_Validate $rest} result]
	    rename _${class}_Validate {}
	    if {$code} {
		global errorInfo
		error "Programming error\nError in validation script\
			for '$name'\n\t$result" $errorInfo
	    }
	    if {[catch {expr {$result && $result}}]} {
		set msg "Programming error\nValidation script return\
			for '$name' is not 0 or 1."
		error $msg $msg
	    }
	    if {!$result} {
		$class Panic "Invalid argument $name=$o"
	    }
	    if {[string match list [lindex $namePair 1]]} {
		break
	    }
	}
	# Process the rest of the command line
	proc _${class}_Process $argSpec $bodyPrefix$body($sswitch)
	set code [catch {eval _${class}_Process $rest} result]
	rename _${class}_Process {}
	if {$code} {
	    error $result "\n$result\n    (command line arguments\
		    '$arg($sswitch)')"
	}
    }

    proc Usage {} {
	set ret "Usage: [Oc_Main GetAppName] "
	set optList [array names body]
	set idx [lsearch -exact $optList $switch]
	set optList [lsort [lreplace $optList $idx $idx]]
	if {[llength $optList]} {
	    append ret {[options]}
	}
	set optcount 0
	set min 0
	foreach namePair $arg($switch) {
	    set name [lindex $namePair 0]
	    set type [lindex $namePair 1]
	    switch -- $type {
		optional {
		    append ret " \[$name"
		    incr optcount
		}
		list {
		    append ret " \[$name ...\]"
		    incr optcount
		    break
		}
		default {
		    if {$optcount} {
			set msg "Programming error\nRequired argument $name\
				follows optional argument(s)."
			error $msg $msg
		    }
		    incr min
		    append ret " $name"
		}
	    }
	}
	if {$optcount + $min < [llength $arg($switch)]} {
	    set msg "Programming error\nList argument $name must be last"
	    error $msg $msg
	}
	while {$optcount} {append ret \]; incr optcount -1}
	if {[llength $arg($switch)]} {
	    append ret "\nwhere"
	}
	foreach namePair $arg($switch) d [lrange $doc($switch) 1 end] {
	    append ret "\n  [format %-13s [lindex $namePair 0]]$d"
	}
	if {![llength $optList]} {
	    return $ret
	}
	append ret "\nAvailable options:"
	foreach opt $optList {
	    append ret "\n  [format %-13s%-13s $switch$opt $arg($opt)]"
	    append ret [lindex $doc($opt) 0]
	    foreach d [lrange $doc($opt) 1 end] {
		append ret "\n                            $d"
	    }
	}
	if {[llength $arg($switch)]} {
	    append ret "\n  [format %-13s%-13s $switch$switch ""]"
	    append ret [lindex $doc($switch) 0]
	}
	return $ret
    }

    proc AvailableOptions {} {
	return [array names body]
    }

    proc Switch {args} {
	if {[llength $args]} {
	    set old $switch
	    set switch [lindex $args 0]
	    set body($switch) $body($old)
	    set valid($switch) $valid($old)
	    set arg($switch) $arg($old)
	    set doc($switch) $doc($old)
	    catch {unset arg($old) body($old) doc($old) valid($old)}
	}
	return $switch
    }

    Constructor {args} {
        return -code error "Can't create instances of $class"
    }

}

